# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail data models
"""

from __future__ import unicode_literals, absolute_import

from .core import Base, ModelBase, uuid_column, getset_factory, GPCType, Setting, Change, Note
from .contact import PhoneNumber, EmailAddress, MailingAddress

from .people import Person, PersonPhoneNumber, PersonEmailAddress, PersonMailingAddress, PersonNote
from .users import Role, Permission, User, UserRole, UserEvent
from .stores import Store, StorePhoneNumber, StoreEmailAddress
from .customers import (Customer, CustomerPhoneNumber, CustomerEmailAddress, CustomerMailingAddress,
                        CustomerGroup, CustomerGroupAssignment, CustomerPerson)

from .org import Department, Subdepartment, Category, Family, ReportCode, DepositLink
from .employees import (Employee, EmployeePhoneNumber, EmployeeEmailAddress,
                        EmployeeStore, EmployeeDepartment, EmployeeHistory)
from .shifts import ScheduledShift, WorkedShift

from .vendors import Vendor, VendorPhoneNumber, VendorEmailAddress, VendorContact
from .products import (Brand, Tax, Product, ProductImage, ProductCode,
                       ProductCost, ProductFutureCost, ProductPrice,
                       ProductInventory, ProductStoreInfo, InventoryAdjustmentReason)
from .purchase import (PurchaseBase, PurchaseItemBase, PurchaseCreditBase,
                       Purchase, PurchaseItem, PurchaseCredit)

from .custorders import CustomerOrder, CustomerOrderItem, CustomerOrderItemEvent

from .messages import Message, MessageRecipient

from .datasync import DataSyncChange
from .labels import LabelProfile
from .bouncer import EmailAttempt, EmailBounce
from .tempmon import TempmonClient, TempmonProbe, TempmonReading
from .upgrades import Upgrade, UpgradeRequirement

from .exports import ExportMixin
from .reports import ReportOutput
from .batch import BatchMixin, BaseFileBatchMixin, FileBatchMixin, BatchRowMixin, ProductBatchRowMixin
from .batch.dynamic import DynamicBatchMixin, ImporterBatch
from .batch.handheld import HandheldBatch, HandheldBatchRow
from .batch.inventory import InventoryBatch, InventoryBatchRow
from .batch.labels import LabelBatch, LabelBatchRow
from .batch.pricing import PricingBatch, PricingBatchRow
from .batch.purchase import PurchaseBatch, PurchaseBatchRow, PurchaseBatchRowClaim, PurchaseBatchCredit
from .batch.vendorcatalog import VendorCatalog, VendorCatalogRow
from .batch.vendorinvoice import VendorInvoice, VendorInvoiceRow



# -*- coding: utf-8 -*-
"""add product.discountable

Revision ID: 3c33f288ad62
Revises: 3544687ae150
Create Date: 2015-02-27 00:28:50.454692

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3c33f288ad62'
down_revision = u'3544687ae150'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product.discountable
    op.add_column('product', sa.Column('discountable', sa.Boolean(), nullable=True))
    product = sa.sql.table('product', sa.sql.column('discountable'))
    op.execute(product.update().values({'discountable': True}))
    op.alter_column('product', 'discountable', nullable=False)
    op.add_column('product_version', sa.Column('discountable', sa.Boolean(), nullable=True))


def downgrade():

    # product.discountable
    op.drop_column('product_version', 'discountable')
    op.drop_column('product', 'discountable')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Dumps
"""

from __future__ import unicode_literals, absolute_import

import six
from sqlalchemy.orm import class_mapper

from rattail.db import model
from rattail.csvutil import UnicodeDictWriter


def dump_data(session, model_class, outfile, order_by=None, progress=None):
    """
    Dump data for a model in CSV format.

    :param session: A SQLAlchemy session instance.

    :param model_class: A model class whose table data will be dumped.

    :param outfile: A file-like object to which the data will be written.

    :param order_by: Optional "ORDER BY" clause element for the query.  If none
       is specified, the model's ``uuid`` attribute will be used.

    :returns: Number of data records written.
    """

    query = session.query(model_class)
    if order_by is not None:
        query = query.order_by(order_by)
    elif hasattr(model_class, 'uuid'):
        query = query.order_by(model_class.uuid)
    count = query.count()
    if not count:
        return 0

    prog = None
    if progress:
        prog = progress("Dumping {} data".format(model_class.__name__), count)

    keys = class_mapper(model_class).columns.keys()

    # TODO: Temp hack until we can overhaul this whole idea.  Currently the
    # only use for the data dump is to do a data diff, which should really not
    # take this timestamp into account.
    if model_class is model.Person:
        keys.remove('modified')

    writer = UnicodeDictWriter(outfile, keys)
    writer.writeheader()

    for i, obj in enumerate(query, 1):
        data = {}
        for key in keys:
            data[key] = six.text_type(getattr(obj, key))
        writer.writerow(data)
        if prog:
            prog.update(i)
    if prog:
        prog.destroy()



# -*- coding: utf-8 -*-
"""add purchase credits

Revision ID: 193c2dcdd14d
Revises: 123af5f6a0bc
Create Date: 2016-12-09 17:03:39.001046

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '193c2dcdd14d'
down_revision = u'123af5f6a0bc'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # purchase_batch_row
    op.add_column('purchase_batch_row', sa.Column('cases_mispick', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('units_mispick', sa.Numeric(precision=10, scale=4), nullable=True))

    # purchase_item
    op.add_column('purchase_item', sa.Column('cases_mispick', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_item', sa.Column('units_mispick', sa.Numeric(precision=10, scale=4), nullable=True))

    # purchase_credit
    op.create_table('purchase_credit',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('store_uuid', sa.String(length=32), nullable=False),
                    sa.Column('vendor_uuid', sa.String(length=32), nullable=False),
                    sa.Column('date_ordered', sa.Date(), nullable=True),
                    sa.Column('date_shipped', sa.Date(), nullable=True),
                    sa.Column('date_received', sa.Date(), nullable=True),
                    sa.Column('invoice_number', sa.String(length=20), nullable=True),
                    sa.Column('invoice_date', sa.Date(), nullable=True),
                    sa.Column('credit_type', sa.String(length=20), nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=60), nullable=False),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('case_quantity', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('cases_shorted', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_shorted', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('expiration_date', sa.Date(), nullable=True),
                    sa.Column('invoice_line_number', sa.Integer(), nullable=True),
                    sa.Column('invoice_case_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('invoice_unit_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('invoice_total', sa.Numeric(precision=7, scale=2), nullable=True),
                    sa.Column('purchase_uuid', sa.String(length=32), nullable=True),
                    sa.Column('item_uuid', sa.String(length=32), nullable=True),
                    sa.Column('status', sa.Integer(), nullable=True),
                    sa.Column('mispick_brand_name', sa.String(length=100), nullable=True),
                    sa.Column('mispick_description', sa.String(length=60), nullable=True),
                    sa.Column('mispick_product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('mispick_size', sa.String(length=255), nullable=True),
                    sa.Column('mispick_upc', rattail.db.types.GPCType(), nullable=True),
                    sa.ForeignKeyConstraint(['item_uuid'], [u'purchase_item.uuid'], name=u'purchase_credit_fk_item'),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'purchase_credit_fk_product'),
                    sa.ForeignKeyConstraint(['purchase_uuid'], [u'purchase.uuid'], name=u'purchase_credit_fk_purchase'),
                    sa.ForeignKeyConstraint(['store_uuid'], [u'store.uuid'], name=u'purchase_credit_fk_store'),
                    sa.ForeignKeyConstraint(['vendor_uuid'], [u'vendor.uuid'], name=u'purchase_credit_fk_vendor'),
                    sa.ForeignKeyConstraint(['mispick_product_uuid'], [u'product.uuid'], name=u'purchase_credit_fk_mispick_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # purchase_batch_credit
    op.create_table('purchase_batch_credit',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('row_uuid', sa.String(length=32), nullable=True),
                    sa.Column('store_uuid', sa.String(length=32), nullable=False),
                    sa.Column('vendor_uuid', sa.String(length=32), nullable=False),
                    sa.Column('date_ordered', sa.Date(), nullable=True),
                    sa.Column('date_shipped', sa.Date(), nullable=True),
                    sa.Column('date_received', sa.Date(), nullable=True),
                    sa.Column('invoice_number', sa.String(length=20), nullable=True),
                    sa.Column('invoice_date', sa.Date(), nullable=True),
                    sa.Column('credit_type', sa.String(length=20), nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=60), nullable=False),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('case_quantity', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('cases_shorted', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_shorted', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('expiration_date', sa.Date(), nullable=True),
                    sa.Column('invoice_line_number', sa.Integer(), nullable=True),
                    sa.Column('invoice_case_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('invoice_unit_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('invoice_total', sa.Numeric(precision=7, scale=2), nullable=True),
                    sa.Column('mispick_brand_name', sa.String(length=100), nullable=True),
                    sa.Column('mispick_description', sa.String(length=60), nullable=True),
                    sa.Column('mispick_product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('mispick_size', sa.String(length=255), nullable=True),
                    sa.Column('mispick_upc', rattail.db.types.GPCType(), nullable=True),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'purchase_batch_credit_fk_product'),
                    sa.ForeignKeyConstraint(['row_uuid'], [u'purchase_batch_row.uuid'], name=u'purchase_batch_credit_fk_row'),
                    sa.ForeignKeyConstraint(['store_uuid'], [u'store.uuid'], name=u'purchase_batch_credit_fk_store'),
                    sa.ForeignKeyConstraint(['vendor_uuid'], [u'vendor.uuid'], name=u'purchase_batch_credit_fk_vendor'),
                    sa.ForeignKeyConstraint(['mispick_product_uuid'], [u'product.uuid'], name=u'purchase_batch_credit_fk_mispick_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # purchase_batch_credit
    op.drop_table('purchase_batch_credit')

    # purchase_credit
    op.drop_table('purchase_credit')

    # purchase_item
    op.drop_column('purchase_item', 'units_mispick')
    op.drop_column('purchase_item', 'cases_mispick')

    # purchase_batch_row
    op.drop_column('purchase_batch_row', 'units_mispick')
    op.drop_column('purchase_batch_row', 'cases_mispick')



# -*- coding: utf-8; -*-
"""add vendor_catalog_row.cost ref

Revision ID: c1bfb033050d
Revises: ccd32c44393f
Create Date: 2018-04-05 22:45:04.638730

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'c1bfb033050d'
down_revision = u'ccd32c44393f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor_catalog_row
    op.add_column('vendor_catalog_row', sa.Column('cost_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'vendor_catalog_row_fk_cost', 'vendor_catalog_row', 'product_cost', ['cost_uuid'], ['uuid'])


def downgrade():

    # vendor_catalog_row
    op.drop_constraint(u'vendor_catalog_row_fk_cost', 'vendor_catalog_row', type_='foreignkey')
    op.drop_column('vendor_catalog_row', 'cost_uuid')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Contact Info
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa

from .core import Base, uuid_column


@six.python_2_unicode_compatible
class PhoneNumber(Base):
    """
    Represents a phone (or fax) number associated with a contactable entity.
    """
    __tablename__ = 'phone'
    __versioned__= {}

    uuid = uuid_column()
    parent_type = sa.Column(sa.String(length=20), nullable=False)
    parent_uuid = sa.Column(sa.String(length=32), nullable=False)
    preference = sa.Column(sa.Integer(), nullable=False)
    type = sa.Column(sa.String(length=15))
    number = sa.Column(sa.String(length=20), nullable=False)

    __mapper_args__ = {'polymorphic_on': parent_type}

    def __str__(self):
        return self.number or ""


@six.python_2_unicode_compatible
class EmailAddress(Base):
    """
    Represents an email address associated with a contactable entity.
    """
    __tablename__ = 'email'
    __versioned__= {}

    uuid = uuid_column()
    parent_type = sa.Column(sa.String(length=20), nullable=False)
    parent_uuid = sa.Column(sa.String(length=32), nullable=False)
    preference = sa.Column(sa.Integer(), nullable=False)
    type = sa.Column(sa.String(length=15))
    address = sa.Column(sa.String(length=255), nullable=False)

    invalid = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether the email address is *known* to be invalid.
    Defaults to NULL, meaning the validity is "not known".
    """)

    __mapper_args__ = {'polymorphic_on': parent_type}

    def __str__(self):
        return self.address or ""


@six.python_2_unicode_compatible
class MailingAddress(Base):
    """
    Represents a physical / mailing address associated with a contactable entity.
    """
    __tablename__ = 'address'
    __versioned__= {}

    uuid = uuid_column()
    parent_type = sa.Column(sa.String(length=20), nullable=False)
    parent_uuid = sa.Column(sa.String(length=32), nullable=False)
    preference = sa.Column(sa.Integer(), nullable=False)
    type = sa.Column(sa.String(length=15), nullable=True)

    street = sa.Column(sa.String(length=100), nullable=True)
    street2 = sa.Column(sa.String(length=100), nullable=True)
    city = sa.Column(sa.String(length=60), nullable=True)
    state = sa.Column(sa.String(length=2), nullable=True)
    zipcode = sa.Column(sa.String(length=10), nullable=True)
    invalid = sa.Column(sa.Boolean(), nullable=True)

    __mapper_args__ = {'polymorphic_on': parent_type}

    def __str__(self):

        if self.street and self.street2:
            street = '{}, {}'.format(self.street, self.street2)
        else:
            street = self.street or ''

        if self.city and self.state:
            city = '{}, {}'.format(self.city, self.state)
        else:
            city = self.city or self.state or ''

        if street and city and self.zipcode:
            text = '{}, {}  {}'.format(street, city, self.zipcode)
        elif street and city:
            text = '{}, {}'.format(street, city)
        elif street and self.zipcode:
            text = '{}  {}'.format(street, self.zipcode)
        elif city and self.zipcode:
            text = '{}  {}'.format(city, self.zipcode)
        else:
            text = self.zipcode or ''

        return text



# -*- coding: utf-8 -*-
"""add batch ids

Revision ID: 064f71d774ad
Revises: 4137e9938122
Create Date: 2016-08-16 19:21:58.543558

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '064f71d774ad'
down_revision = u'4137e9938122'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():
    from sqlalchemy.schema import CreateSequence

    # batch_id_seq
    op.execute(CreateSequence(batch_id_seq))

    # vendor_catalog
    op.add_column('vendor_catalog', sa.Column('id', sa.Integer(), batch_id_seq, nullable=True))
    vendor_catalog = sa.sql.table('vendor_catalog', sa.sql.column('id'))
    op.execute(vendor_catalog.update().values({'id': batch_id_seq.next_value()}))
    op.alter_column('vendor_catalog', 'id', nullable=False)

    # vendor_invoice
    op.add_column('vendor_invoice', sa.Column('id', sa.Integer(), batch_id_seq, nullable=True))
    vendor_invoice = sa.sql.table('vendor_invoice', sa.sql.column('id'))
    op.execute(vendor_invoice.update().values({'id': batch_id_seq.next_value()}))
    op.alter_column('vendor_invoice', 'id', nullable=False)


def downgrade():

    # vendor_invoice
    op.drop_column('vendor_invoice', 'id')

    # vendor_catalog
    op.drop_column('vendor_catalog', 'id')



# -*- coding: utf-8 -*-
"""add product.notes

Revision ID: 471299288da9
Revises: 234c79420312
Create Date: 2017-02-27 15:33:20.114167

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '471299288da9'
down_revision = u'234c79420312'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # product
    op.add_column('product', sa.Column('ingredients', sa.Text(), nullable=True))
    op.add_column('product', sa.Column('notes', sa.Text(), nullable=True))


def downgrade():

    # product
    op.drop_column('product', 'notes')
    op.drop_column('product', 'ingredients')



# -*- coding: utf-8 -*-
"""add custorders

Revision ID: d6ca20fe9452
Revises: 09ed5bdc99f1
Create Date: 2017-02-16 20:53:48.730360

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'd6ca20fe9452'
down_revision = u'09ed5bdc99f1'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # custorder
    op.create_table('custorder',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=True),
                    sa.Column('customer_uuid', sa.String(length=32), nullable=True),
                    sa.Column('person_uuid', sa.String(length=32), nullable=True),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['customer_uuid'], [u'customer.uuid'], name=u'custorder_fk_customer'),
                    sa.ForeignKeyConstraint(['person_uuid'], [u'person.uuid'], name=u'custorder_fk_person'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # custorder_item
    op.create_table('custorder_item',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('order_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('product_brand', sa.String(length=100), nullable=True),
                    sa.Column('product_description', sa.String(length=60), nullable=True),
                    sa.Column('product_size', sa.String(length=30), nullable=True),
                    sa.Column('product_unit_of_measure', sa.String(length=4), nullable=False),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('case_quantity', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('cases_ordered', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_ordered', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('product_unit_cost', sa.Numeric(precision=9, scale=5), nullable=True),
                    sa.Column('unit_price', sa.Numeric(precision=8, scale=3), nullable=True),
                    sa.Column('discount_percent', sa.Numeric(precision=5, scale=3), nullable=False),
                    sa.Column('total_price', sa.Numeric(precision=8, scale=3), nullable=True),
                    sa.Column('paid_amount', sa.Numeric(precision=8, scale=3), nullable=False),
                    sa.Column('payment_transaction_number', sa.String(length=8), nullable=True),
                    sa.Column('status_code', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(['order_uuid'], [u'custorder.uuid'], name=u'custorder_item_fk_order'),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'custorder_item_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # custorder_item_event
    op.create_table('custorder_item_event',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('item_uuid', sa.String(length=32), nullable=False),
                    sa.Column('type_code', sa.Integer(), nullable=False),
                    sa.Column('occurred', sa.DateTime(), nullable=False),
                    sa.Column('user_uuid', sa.String(length=32), nullable=False),
                    sa.Column('note', sa.Text(), nullable=True),
                    sa.ForeignKeyConstraint(['item_uuid'], [u'custorder_item.uuid'], name=u'custorder_item_event_fk_item'),
                    sa.ForeignKeyConstraint(['user_uuid'], [u'user.uuid'], name=u'custorder_item_event_fk_user'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # custorder*
    op.drop_table('custorder_item_event')
    op.drop_table('custorder_item')
    op.drop_table('custorder')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Console Stuff
"""

from __future__ import unicode_literals, absolute_import

import os
import sys
import json
import socket

import six
from progress.bar import Bar


class ProgressBase(object):
    """
    Base class for all progress implementations.
    """

    def __init__(self, message, maximum):
        self.message = message
        self.maximum = maximum

    def update(self, value):
        return True

    def finish(self):
        pass

    def destroy(self):
        self.finish()


class ConsoleProgress(ProgressBase):
    """
    Provides a console-based progress bar.
    """

    def __init__(self, message, maximum, stdout=None):
        self.stdout = stdout or sys.stderr
        self.stdout.write("\n{}...\n".format(message))
        self.bar = Bar(None, max=maximum, width=70,
                       suffix='%(index)d/%(max)d %(percent)d%% ETA %(eta)ds')

    def update(self, value):
        self.bar.next()
        return True

    def finish(self):
        self.bar.finish()


class FileProgress(ProgressBase):
    """
    Progress indicator which writes progress info to the given file object.
    """

    def __init__(self, path):
        self.path = path

    def __call__(self, message, maximum):
        self.message = message
        self.maximum = maximum
        self.value = 0
        self.write_info()
        return self

    def write_info(self):
        info = {
            'message': self.message,
            'maximum': self.maximum,
            'value': self.value,
        }
        with open(self.path, 'wt') as f:
            f.write(json.dumps(info))
            f.flush()

    def update(self, value):
        self.value = value
        self.write_info()
        return True

    def finish(self):
        if os.path.exists(self.path):
            os.remove(self.path)


class SocketProgress(ProgressBase):
    """
    Progress indicator which writes progress info to the given socket object.
    """

    def __init__(self, host, port, suffix=None):
        self.host = host
        self.port = port
        if suffix:
            self.suffix = suffix
        else:
            self.suffix = "\n\n.".encode('utf_8')

    def __call__(self, message, maximum):
        self.connection = socket.create_connection((self.host, self.port))
        self.message = message
        self.maximum = maximum
        self.value = 0
        self.write_progress()
        return self

    def write_data(self, data):
        self.connection.send(json.dumps(data))
        self.connection.send(self.suffix)

    def write_progress(self):
        self.write_data({
            'message': self.message,
            'maximum': self.maximum,
            'value': self.value,
        })

    def update(self, value):
        self.value = value
        self.write_progress()
        return True

    def finish(self):
        self.write_data({
            'process_complete': True,
        })
        self.connection.close()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Batch-related commands
"""

from __future__ import unicode_literals, absolute_import

import logging

from rattail.commands.core import Subcommand
from rattail.progress import SocketProgress
from rattail.time import make_utc


log = logging.getLogger(__name__)


class BatchAction(Subcommand):
    """
    Base class for commands which invoke a handler to act on a batch.
    """

    def add_parser_args(self, parser):
        parser.add_argument('batch_type',
                            help="Type key of batch to be populated, e.g. 'vendor_catalog'.")
        parser.add_argument('batch_uuid',
                            help="UUID of the batch to be populated.")
        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        from rattail.batch import get_batch_handler

        handler = get_batch_handler(self.config, args.batch_type)
        if not handler:
            raise RuntimeError("Could not locate handler for batch type: {}".format(args.batch_type))

        session = self.make_session()
        user = self.get_runas_user(session)

        batch = session.query(handler.batch_model_class).get(args.batch_uuid)
        if not batch:
            raise RuntimeError("Batch of type '{}' not found: {}".format(args.batch_type, args.batch_uuid))

        success = self.action(handler, batch, user)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        elif success:
            session.commit()
            log.info("transaction was committed")
        else:
            session.rollback()
            log.warning("action failed, so transaction was rolled back")
        session.close()


class PopulateBatch(BatchAction):
    """
    Populate initial data for a batch
    """
    name = 'populate-batch'
    description = __doc__.strip()

    def action(self, handler, batch, user):
        return handler.do_populate(batch, user, progress=self.progress)


class RefreshBatch(BatchAction):
    """
    Refresh data for a batch
    """
    name = 'refresh-batch'
    description = __doc__.strip()

    def action(self, handler, batch, user):
        return handler.do_refresh(batch, user, progress=self.progress)


class ExecuteBatch(BatchAction):
    """
    Execute a batch
    """
    name = 'execute-batch'
    description = __doc__.strip()

    def action(self, handler, batch, user):
        return handler.do_execute(batch, user, progress=self.progress)
